/*
    Ishtar - A network transparent read/write interface with
    probing and enumeration support.
    service part : network transparent access to generic services
    Copyright (C) 2003-2005 Stephane Magnenat <nct@ysagoon.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "services.h"
#include <sstream>
#include <list>

/*!	\file services.cpp
	\brief Implementation of Ishtar services classes.
*/

namespace Ishtar
{
	// Value Vector
	// ============
	
	ValueVector *ValueVector::newOfType(Type type, Size size)
	{
		switch (type)
		{
			case ValueVector::INT:
			return new ValueVectorTemplate<Int32>(size);
			
			case ValueVector::UINT:
			return new ValueVectorTemplate<UInt32>(size);
			
			case ValueVector::SHORT:
			return new ValueVectorTemplate<Int16>(size);
			
			case ValueVector::USHORT:
			return new ValueVectorTemplate<UInt16>(size);
			
			case ValueVector::CHAR:
			return new ValueVectorTemplate<Int8>(size);
			
			case ValueVector::UCHAR:
			return new CharValueVector(size);
			
			case ValueVector::FLOAT:
			return new ValueVectorTemplate<float>(size);
			
			case ValueVector::DOUBLE:
			return new ValueVectorTemplate<double>(size);
			
			case ValueVector::BOOL:
			return new ValueVectorTemplate<bool>(size);
	
			default:
			assert(false);
			return NULL;
		}
	}
	
	ValueVector *ValueVector::deserialize(InputStream *stream)
	{
		Type type = static_cast<Type>(stream->readUInt32());
		ValueVector *vv = newOfType(type);
		vv->deserializeSpecified(stream);
		return vv;
	}
	
	CharValueVector::CharValueVector(const std::string &s)
	:ValueVectorTemplate<UInt8>(s.size())
	{
		std::copy(s.begin(), s.end(), value.begin());
	}
	
	CharValueVector::CharValueVector(const char *s)
	:ValueVectorTemplate<UInt8>(strlen(s))
	{
		for (unsigned i=0; i<strlen(s); i++)
		{
			value[i]=s[i];
		}
	}
	
	CharValueVector::operator std::string() const
	{
		std::string s;
		s.resize(value.size());
		std::copy(value.begin(), value.end(), s.begin());
		return s;
	}
	
	// Service Description
	// ===================
	
	ServiceDescription::ServiceDescription()
	{
		type=ValueVector::INT;
		flags=0;
		length=0;
		mins=NULL;
		maxs=NULL;
	}
	
	ServiceDescription::~ServiceDescription()
	{
		if (mins)
			delete mins;
		if (maxs)
			delete maxs;
	}
	
	
	void ServiceDescription::serialize(OutputStream *stream)
	{
		name.serializeSpecified(stream);
	
		stream->writeUInt32(type);
		stream->writeUInt32(flags);
		stream->writeUInt32(length);
	
		if ((flags & NAMED_VALUES) && (length>0))
		{
			for (unsigned i=0; i<length; i++)
			{
				subnames[i].serializeSpecified(stream);
			}
		}
	
		if ((flags & CONSTRAINT_VALUES) && (length>0))
		{
			mins->serialize(stream);
			maxs->serialize(stream);
		}
	}
	
	void ServiceDescription::deserialize(InputStream *stream)
	{
		name.deserializeSpecified(stream);
	
		type = static_cast<ValueVector::Type>(stream->readUInt32());
		flags = stream->readUInt32();
		length = stream->readUInt32();
	
		if ((flags & NAMED_VALUES) && (length>0))
		{
			subnames.resize(length);
			for (unsigned i=0; i<length; i++)
			{
				subnames[i].deserializeSpecified(stream);
			}
		}
		else
		{
			subnames.clear();
		}
	
		if (mins)
		{
			delete mins;
			mins = NULL;
		}
		if (maxs)
		{
			delete maxs;
			maxs = NULL;
		}
	
		if ((flags & CONSTRAINT_VALUES) && (length>0))
		{
			mins = ValueVector::deserialize(stream);
			maxs = ValueVector::deserialize(stream);
		}
	}
	
	// ServicesContainer
	// =================
	
	ServicesContainer::ServicesContainer()
	{
		maxUID = 0;
	}
	
	ServicesContainer::~ServicesContainer()
	{
		for (std::map<UInt32, Service *>::iterator it=services.begin(); it!=services.end(); ++it)
		{
			delete (it->second);
		}
	}
	
	void ServicesContainer::addService(Service *s)
	{
		reverseServices[s] = maxUID;
		services[maxUID++] = s;
		namedServices[s->description.name] = s;
	}
	
	void ServicesContainer::eraseService(Service *s)
	{
		std::map<Service *, UInt32>::iterator it = reverseServices.find(s);
		if (it != reverseServices.end())
		{
			services.erase(it->second);
			reverseServices.erase(it);
			namedServices.erase(s->description.name);
			delete s;
		}
	}
	
	Service *ServicesContainer::findService(const std::string &name)
	{
		std::map<std::string, Service *>::iterator it = namedServices.find(name);
		if (it !=  namedServices.end())
		{
			return it->second;
		}
		else
		{
			return NULL;
		}
	}
	
	// ServicesServer
	// =================
	
	ServicesServer::ServicesServer(UInt16 port) :
		NetworkServer(port)
	{	
	}
	
	ServicesServer::~ServicesServer()
	{
	}
	
	void ServicesServer::incomingData(Socket *socket)
	{
		//std::cout << "Call received" << std::endl;
		// Read call
		Call *call = Call::deserialize(socket);
	
		// Execute call & delete it
		Answer *answer = call->execute(this);
	
		// delete
		delete call;
	
		// Return answer & delete it
		if (answer)
		{
			answer->serialize(socket);
			socket->flush();
	
			// delete
			ValuesAnswer *valuesAnswer = dynamic_cast<ValuesAnswer *>(answer);
			if (valuesAnswer)
				valuesAnswer->deleteValues();
			delete answer;
		}
	}
	
	void ServicesServer::connectionClosed(Socket *socket)
	{
		// Connection has been close, so is the life
	}
	
	void ServicesServer::incomingConnection(Socket *socket)
	{
		Call *call = Call::deserialize(socket);
		HelloCall *hello = dynamic_cast<HelloCall *>(call);
		if (hello)
		{
			if (hello->nodeType == HelloCall::CLIENT)
			{
				//std::cout << "New client connection from " << socket->getRemoteName() << std::endl;
				if (hello->version != Ishtar::PROTOCOL_VERSION)
				{
					std::cerr << "Client " << socket->getRemoteName() << " has wrong Ishtar Protocol Version (" << hello->version << " instead of " << Ishtar::PROTOCOL_VERSION << ")" << std::endl;
				}
				else
				{
					delete call;
					return;
				}
			}
		}
		
		delete call;
		std::cerr << "New connection error : first token was not a valid hello one from a client" << std::endl;
		socket->disconnect();
	}
	
	
	// Calls and answers
	// =================
	
	Answer *Answer::deserialize(InputStream *stream)
	{
		Type type = static_cast<Type>(stream->readUInt32());
		switch (type)
		{
			case Answer::ERROR:
			{
				ErrorAnswer *answer = new ErrorAnswer;
				answer->deserializeSpecified(stream);
				return answer;
			}
			case Answer::SERVICE_LIST:
			{
				ServiceAnswer *answer = new ServiceAnswer;
				answer->deserializeSpecified(stream);
				return answer;
			}
			case Answer::VALUES:
			{
				ValuesAnswer *answer = new ValuesAnswer;
				answer->deserializeSpecified(stream);
				return answer;
			}
			default:
			assert(false);
		}
	}
	
	Call *Call::deserialize(InputStream *stream)
	{
		Type type = static_cast<Type>(stream->readUInt32());
		switch (type)
		{
			case Call::HELLO:
			{
				HelloCall *call = new HelloCall;
				call->deserializeSpecified(stream);
				return call;
			}
			case Call::GET_SERVICE_LIST:
			{
				GetServicesCall *call = new GetServicesCall;
				call->deserializeSpecified(stream);
				return call;
			}
			case Call::GET_VALUES:
			{
				GetValuesCall *call = new GetValuesCall;
				call->deserializeSpecified(stream);
				return call;
			}
			case Call::SET_VALUES:
			{
				SetValuesCall *call = new SetValuesCall;
				call->deserializeSpecified(stream);
				return call;
			}
			default:
			assert(false);
		}
	}
	
	// HelloCall
	
	void HelloCall::serialize(OutputStream *stream)
	{
		stream->writeUInt32(HELLO);
		stream->writeUInt32(nodeType);
		stream->writeUInt32(version);
	}
	
	void HelloCall::deserializeSpecified(InputStream *stream)
	{
		nodeType = static_cast<HelloCall::NodeType>(stream->readUInt32());
		version = stream->readUInt32();
	}
	
	// Error answer
	
	void ErrorAnswer::serialize(OutputStream *stream)
	{
		stream->writeUInt32(ERROR);
		stream->writeUInt32(errorType);
	}
	
	void ErrorAnswer::deserializeSpecified(InputStream *stream)
	{
		errorType = static_cast<ErrorAnswer::ErrorTypes>(stream->readUInt32());
	}
	
	// GetServicesCall
	
	Answer *GetServicesCall::execute(ServicesContainer *registrar)
	{
		ServiceAnswer *answer = new ServiceAnswer;
		for (std::map<UInt32, Service *>::iterator it = registrar->services.begin(); it != registrar->services.end(); ++it)
		{
			answer->descriptions[it->first] = &(it->second->description);
		}
		return answer;
	}
	
	// ServiceAnswer
	
	ServiceAnswer::~ServiceAnswer()
	{
		/* NOTE:
		We do not delete services descriptions because they are
		just pointers to services */
	}
	
	void ServiceAnswer::deleteDescriptions(void)
	{
		for (std::map<UInt32, ServiceDescription *>::iterator it=descriptions.begin(); it!=descriptions.end(); ++it)
		{
			delete (it->second);
		}
	}
	
	void ServiceAnswer::serialize(OutputStream *stream)
	{
		stream->writeUInt32(SERVICE_LIST);
		stream->writeUInt32(descriptions.size());
		for (std::map<UInt32, ServiceDescription *>::iterator it=descriptions.begin(); it!=descriptions.end(); ++it)
		{
			stream->writeUInt32(it->first);
			it->second->serialize(stream);
		}
	}
	
	void ServiceAnswer::deserializeSpecified(InputStream *stream)
	{
		assert(descriptions.size()==0);
		unsigned size = stream->readUInt32();
		for (unsigned i=0; i<size; i++)
		{
			UInt32 id;
			ServiceDescription *sd;
	
			id = stream->readUInt32();
			sd = new ServiceDescription;
			sd->deserialize(stream);
			descriptions[id] = sd;
		}
	}
	
	// GetValuesCall
	
	void GetValuesCall::serialize(OutputStream *stream)
	{
		stream->writeUInt32(GET_VALUES);
		stream->writeInt32(requestID);
		stream->writeUInt32(services.size());
		for (unsigned i=0; i<services.size(); i++)
		{
			stream->writeInt32(services[i]);
		}
	}
	
	void GetValuesCall::deserializeSpecified(InputStream *stream)
	{
		requestID = stream->readInt32();
		assert(services.size()==0);
		unsigned size = stream->readUInt32();
		services.resize(size);
		for (unsigned i=0; i<size; i++)
		{
			services[i] = stream->readUInt32();
		}
	}
	
	Answer *GetValuesCall::execute(ServicesContainer *registrar)
	{
		ValuesAnswer *answer = new ValuesAnswer();
	
		answer->requestID=requestID;
	
		for (unsigned i=0; i<services.size(); i++)
		{
			UInt32 id = services[i];
	
			std::map<UInt32, Service *>::iterator it = registrar->services.find(id);
			if (it != registrar->services.end())
			{
				ValueVector *v = it->second->getValues();
				answer->values[id] = v;
			}
			else
			{
				std::cerr << "Service " << id << " does not exists" << std::endl;
				delete answer;
				return new ErrorAnswer(ErrorAnswer::SERVICE_DOES_NOT_EXISTS);
			}
		}
		return answer;
	}
	
	// ValuesAnswer
	
	ValuesAnswer::~ValuesAnswer()
	{
		/* NOTE:
		We do not delete values because they are
		just pointers to ValueVector and aren't memcopied
		for the sake of optimisation */
	}
	
	void ValuesAnswer::deleteValues(void)
	{
		for (std::map<UInt32, ValueVector *>::iterator it=values.begin(); it!=values.end(); ++it)
		{
			delete (it->second);
		}
	}
	
	void ValuesAnswer::serialize(OutputStream *stream)
	{
		stream->writeUInt32(VALUES);
		stream->writeInt32(requestID);
		stream->writeUInt32(values.size());
		for (std::map<UInt32, ValueVector *>::iterator it=values.begin(); it!=values.end(); ++it)
		{
			stream->writeUInt32(it->first);
			it->second->serialize(stream);
		}
	}
	
	void ValuesAnswer::deserializeSpecified(InputStream *stream)
	{
		assert(values.size()==0);
		requestID = stream->readInt32();
		unsigned size = stream->readUInt32();
		for (unsigned i=0; i<size; i++)
		{
			unsigned id;
			ValueVector *vv;
	
			id = stream->readUInt32();
			vv = ValueVector::deserialize(stream);
			values[id] = vv;
		}
	}
	
	// SetValuesCall
	
	SetValuesCall::~SetValuesCall()
	{
		/* NOTE:
		We do not delete values because they are
		just pointers to ValueVector and aren't memcopied
		for the sake of optimisation */
	}
	
	void SetValuesCall::deleteValues(void)
	{
		for (std::map<unsigned, ValueVector *>::iterator it=values.begin(); it!=values.end(); ++it)
		{
			delete (it->second);
		}
	}
	
	void SetValuesCall::serialize(OutputStream *stream)
	{
		stream->writeUInt32(SET_VALUES);
		stream->writeUInt32(values.size());
		for (std::map<unsigned, ValueVector *>::iterator it=values.begin(); it!=values.end(); ++it)
		{
			stream->writeUInt32(it->first);
			it->second->serialize(stream);
		}
	}
	
	void SetValuesCall::deserializeSpecified(InputStream *stream)
	{
		assert(values.size()==0);
		unsigned size = stream->readUInt32();
		for (unsigned i=0; i<size; i++)
		{
			unsigned id;
			ValueVector *vv;
	
			id = stream->readUInt32();
			vv = ValueVector::deserialize(stream);
			values[id] = vv;
		}
	}
	
	Answer *SetValuesCall::execute(ServicesContainer *registrar)
	{
		for (std::map<UInt32, ValueVector *>::const_iterator it=values.begin(); it!=values.end(); ++it)
		{
			UInt32 id = it->first;
			
			std::map<UInt32, Service *>::iterator it2 = registrar->services.find(id);
			if (it2 != registrar->services.end())
			{
				Service *s = it2->second;
				
				if (s->description.flags & ServiceDescription::READ_ONLY)
				{
					std::cerr << "Service " << id << " is read-only, can't write " << std::endl;
					continue;
				}
				
				s->setValues(it->second);
			}
			else
			{
				std::cerr << "Service " << id << " does not exists" << std::endl;
				return new ErrorAnswer(ErrorAnswer::SERVICE_DOES_NOT_EXISTS);
			}
		}
	
		// we do not need the values any more
		deleteValues();
		return NULL;
	}
}
